import os
import sys
import time
import asyncio
import tempfile
import traceback


@asyncio.coroutine
def check_for_newerfile(future, lockfile, interval):
    exists = os.path.exists
    mtime = lambda p: os.stat(p).st_mtime
    files = dict()

    for module in list(sys.modules.values()):
        path = getattr(module, '__file__', '')
        if path[-4:] in ('.pyo', '.pyc'): path = path[:-1]
        if path and exists(path): files[path] = mtime(path)

    @asyncio.coroutine
    def reccur():
        status = None
        yield from asyncio.sleep(interval)

        if not exists(lockfile) or mtime(lockfile) < time.time() - interval - 5:
            status = 'error'

        for path, lmtime in list(files.items()):
            if not exists(path) or mtime(path) > lmtime:
                status = 'reload'
                print('Pending reload...')
                break

        if status:
           future.set_result(status)
        else:
           asyncio.async(reccur())

    asyncio.async(reccur())


def reloader_opt(to_call, reloader, interval):
    if reloader and not os.environ.get('PROCESS_CHILD'):
        import subprocess
        lockfile = None
        try:
            fd, lockfile = tempfile.mkstemp(prefix='process.', suffix='.lock')
            os.close(fd)  # We only need this file to exist. We never write to it
            while os.path.exists(lockfile):
                args = [sys.executable] + sys.argv
                environ = os.environ.copy()
                environ['PROCESS_CHILD'] = 'true'
                environ['PROCESS_LOCKFILE'] = lockfile
                p = subprocess.Popen(args, env=environ)
                while p.poll() is None:  # Busy wait...
                    os.utime(lockfile, None)  # I am alive!
                    time.sleep(interval)
                if p.poll() != 3:
                    if os.path.exists(lockfile): os.unlink(lockfile)
                    sys.exit(p.poll())
        except KeyboardInterrupt:
            pass
        finally:
            if os.path.exists(lockfile):
                os.unlink(lockfile)
        return

    try:
        if reloader:
            lockfile = os.environ.get('PROCESS_LOCKFILE')

            future = asyncio.Future()
            asyncio.async(check_for_newerfile(future, lockfile, interval))

            def done(future):
                # Stop event loop
                loop = asyncio.get_event_loop()
                if loop.is_running() and future.result() != 'error':
                    loop.stop()

            future.add_done_callback(done)

            to_call()

            if future.done() and future.result() == 'reload':
                sys.exit(3)

        else:
            to_call()

    except KeyboardInterrupt:
        pass
    except (SystemExit, MemoryError):
        raise
    except:
        if not reloader: raise
        traceback.print_exc()
        time.sleep(interval)
        sys.exit(3)




import asyncio

class Event:
    def __init__(self):
        self.handlers = set()

    def handle(self, handler):
        self.handlers.add(handler)
        return self

    def unhandle(self, handler):
        try:
            self.handlers.remove(handler)
        except:
            raise ValueError("Handler is not handling this event, so cannot unhandle it.")
        return self

    @asyncio.coroutine
    def fire(self, *args, **kargs):
        for handler in self.handlers:
            yield from handler(*args, **kargs)

    def getHandlerCount(self):
        return len(self.handlers)

    __iadd__ = handle
    __isub__ = unhandle
    __call__ = fire
    __len__  = getHandlerCount


#!/usr/bin/env python

# # Copyright  : (C) 2014 by MHComm. All rights reserved
#
# Name       : conf.py
"""
   Summary    :  TBD...
"""
from __future__ import absolute_import

__author__ = "jeremie"
__copyright__ = "(C) 2016 by MHComm. All rights reserved"
__email__ = "info@mhcomm.fr"

import sys
import importlib
import traceback
import os
import pypeman.default_settings as default_settings
import logging
import logging.config

NOT_FOUND = object() # sentinel object


class Settings():
    """ pypeman projects settings. rather similiar implemantionts to django.conf.settings """

    def __init__(self):
        self.__dict__['_settings_mod'] = None
        self.__dict__['SETTINGS_MODULE'] = os.environ.get('PYPEMAN_SETTINGS_MODULE', 'settings')

        # TODO Not using lazy setting import as it cause trouble with logging for now
        # TODO is lazy loading really necessary ?
        self._init_settings()

    def _init_settings(self):
        try:
            settings_module = self.__dict__['SETTINGS_MODULE']
            settings_mod = self.__dict__['_settings_mod'] = importlib.import_module(settings_module)
        except:
            print("Can't import 'settings' module !", file=sys.stderr)
            print(traceback.format_exc(), file=sys.stderr)
            sys.exit(-1)

        # populate entire dict with values. helpful e.g. for ipython tab completion
        default_vals = [ (key, val) for (key, val) in default_settings.__dict__.items()
                if 'A' <= key[0] <= 'Z']
        self.__dict__.update(default_vals)

        mod_vals = [ (key, val) for (key, val) in settings_mod.__dict__.items()
                if 'A' <= key[0] <= 'Z']
        self.__dict__.update(mod_vals)

        logging.config.dictConfig(self.__dict__['LOGGING'])

    def __getattr__(self, name):
        """ lazy getattr. first access imports and populates settings """
        if name in self.__dict__:
            return self.__dict__[name]

        if not self.__dict__['_settings_mod']:
            self._init_settings()

        return self.__dict__[name]

    def __setattr__(self, name, value):
        """ make sure nobody tries to modify settings manually """
        if name in self.__dict__:
             self.__dict__[name] = value
        else:
             print(name,value)
             raise Exception("Settings are not editable !")


settings = Settings()



import time
import logging

logger = logging.getLogger(__name__)

class MapItem:
    def __init__(self, old=None, new=None, default=None, transform=None):
        self.old = old
        self.new = new if new else old
        if callable(transform):
            self.transform = transform
        else:
            self.transform = lambda x, y:x
        self.default = default

    def conv(self, oldDict, newDict, msg):
        value = oldDict
        if self.old:
            for part in self.old.split('.'):
                if not value == None:
                    value = value.get(part)

            value = self.transform(value, msg)

        if (not self.old or not value) and self.default is not None:
            value = self.default

        dest = newDict
        parts = self.new.split('.')
        for part in parts[:-1]:
            if dest.get(part) == None:
                dest[part] = {}
            dest = dest[part]

        dest[parts[-1]] = value


class JoinMapItem(MapItem):
    def __init__(self, old, new, sep=''):
        self.sep = sep
        super().__init__(old, new)
    def conv(self, oldDict, newDict, msg):
        values = []
        for value in self.old:
            if oldDict.get(value):
                values.append(oldDict[value])
        strg = self.sep.join(values)
        newDict[self.new] = strg



#!/usr/bin/env python

DEBUG = False   # bool. can be set by env var PYPEMAN_DEBUG (0|1|true|false) or pypeman cmd args
TESTING = False # bool. can be set by env var PYPEMAN_TESTING (0|1|true|false) pypeman cmd args

HTTP_ENDPOINT_CONFIG = ['0.0.0.0', '8080']

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            #'format': '%(levelname)s %(asctime)s %(name)s %(module)s %(process)d %(thread)d %(message)s'
            'format': '%(levelname)s %(asctime)s %(name)s %(module)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level':'DEBUG',
            'class':'logging.StreamHandler',
            'formatter': 'verbose'
        }
    },

    'loggers': {
        # root loggers
        '': {
            'level': 'INFO',
            'handlers': ['console'],
        },
    }
}



import asyncio
import sys
import warnings

all = []

# used to share external dependencies
ext = {}

class BaseEndpoint:
    dependencies = [] # List of module requirements

    def __init__(self):
        all.append(self)

    def requirements(self):
        """ List dependencies of modules if any """
        return self.dependencies

    def import_modules(self):
        """ Use this method to import specific external modules listed in dependencies """
        pass


class HTTPEndpoint(BaseEndpoint):
    dependencies = ['aiohttp']

    def __init__(self, adress='127.0.0.1', port='8080'):
        super().__init__()
        self._app = None
        self.address = adress
        self.port = port

    def import_modules(self):
        if 'aiohttp_web' not in ext:
            from aiohttp import web

            ext['aiohttp_web'] = web

    def add_route(self,*args, **kwargs):
        if not self._app:
            loop = asyncio.get_event_loop()
            self._app = ext['aiohttp_web'].Application(loop=loop)
        # TODO route should be added later
        self._app.router.add_route(*args, **kwargs)

    @asyncio.coroutine
    def start(self):
        if self._app is not None:
            loop = asyncio.get_event_loop()
            srv = yield from loop.create_server(self._app.make_handler(), self.address, self.port)
            print("Server started at http://{}:{}".format(self.address, self.port))
            return srv
        else:
            print("No HTTP route.")

class MLLPProtocol(asyncio.Protocol):
    """
    Minimal Lower-Layer Protocol (MLLP) takes the form:
        <VT>[HL7 Message]<FS><CR>
    References:
    .. [1] http://www.hl7standards.com/blog/2007/05/02/hl7-mlp-minimum-layer-protocol-defined/
    .. [2] http://www.hl7standards.com/blog/2007/02/01/ack-message-original-mode-acknowledgement/
    """

    def __init__(self, handler):
        super().__init__()
        self._buffer = b''
        self.start_block = b'\x0b'  # <VT>, vertical tab
        self.end_block = b'\x1c'  # <FS>, file separator
        self.carriage_return = b'\x0d'  # <CR>, \r
        self.handler = handler

    def connection_made(self, transport):
        """
        Called when a connection is made.
        The argument is the transport representing the pipe connection.
        To receive data, wait for data_received() calls.
        When the connection is closed, connection_lost() is called.
        """
        self.transport = transport

    def process_response(self, future):
        self.writeMessage(future.result())

    def data_received(self, data):
        """
        Called when some data is received.
        The argument is a bytes object.
        """
        # try to find a complete message(s) in the combined the buffer and data
        messages = (self._buffer + data).split(self.end_block)
        # whatever is in the last chunk is an uncompleted message, so put back
        # into the buffer
        self._buffer = messages.pop(-1)

        for raw_message in messages:
            # strip the rest of the MLLP shell from the HL7 message
            raw_message = raw_message.strip(self.start_block + self.carriage_return)

            print(repr(raw_message))

            # only pass messages with data
            if len(raw_message) > 0:
                result = asyncio.async(self.handler(raw_message))
                result.add_done_callback(self.process_response)


    def writeMessage(self, message):
        # convert back to a byte string
        # wrap message in payload container
        self.transport.write(self.start_block + message + self.end_block + self.carriage_return)


    def connection_lost(self, exc):
        """
        Called when the connection is lost or closed.
        The argument is an exception object or None (the latter
        meaning a regular EOF is received or the connection was
        aborted or closed).
        """
        super().connection_lost(exc)


class MLLPEndpoint(BaseEndpoint):
    dependencies = ['hl7']

    def __init__(self, address='127.0.0.1', port='2100', encoding='utf-8'):
        super().__init__()
        self.handlers = []
        self.address = address
        self.port = port

        if encoding != 'utf-8':
            warnings.warn("MLLPEndpoint 'encoding' parameters is deprecated", DeprecationWarning)
        self.encoding = encoding

    def import_modules(self):
        if 'hl7' not in ext:
            import hl7
            ext['hl7'] = hl7

    def set_handler(self, handler):
        self.handler = handler

    @asyncio.coroutine
    def start(self):
        if self.handler:
            loop = asyncio.get_event_loop()
            srv = yield from loop.create_server(lambda: MLLPProtocol(self.handler), self.address, self.port)
            print("MLLP server started at http://{}:{}".format(self.address, self.port))
            return srv
        else:
            print("No MLLP handlers.")





import logging
from pypeman.message import Message
import os
import re
from collections import OrderedDict

logger = logging.getLogger("pypeman.store")

DATE_FORMAT = '%Y%m%d_%H%M'


class MessageStoreFactory():
    """ Message store factory class can generate Message store instance for specific store_id. """

    def get_store(self, store_id):
        """
        :param store_id: identifier of corresponding message store.
        :return: A MessageStore corresponding to correct store_id.
        """

class MessageStore():
    """ A MessageStore keep an history of processed messages. Mainly used in channels. """

    def store(self, msg):
        """
        Store a message in the store.
        :param msg: The message to store.
        :return: Id for this specific message.
        """

    def change_message_state(self, id, new_state):
        """
        Change the `id` message state.
        :param id: Message specific store id.
        :param new_state: Target state.
        :return: -
        """

    def get(self, id):
        """
        Return one message corresponding to given `id` with his status.
        :param id: Message id. Message store dependant.
        :return: A dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """

    def search(self, order_by='timestamp'):
        """
        Return a list of message with store specific `id` and processed status.
        :param start: First element.
        :param count: Count of elements since first element.
        :param order_by: Message order. Allowed values : ['timestamp', 'status'].
        :return: A list of dict `{'id':<message_id>, 'state': <message_state>, 'message': <message_object>}`.
        """


class NullMessageStoreFactory(MessageStoreFactory):
    """ Return an NullMessageStore that do nothing at all. """
    def get_store(self, store_id):
        return NullMessageStore()


class NullMessageStore(MessageStore):
    """ For testing purpose """

    def store(self, msg):
        return None

    def get(self, id):
        return None

    def search(self, order_by='timestamp'):
        return None


class FakeMessageStoreFactory(MessageStoreFactory):
    """ Return an Fake message store """
    def get_store(self, store_id):
        return FakeMessageStore()


class FakeMessageStore(MessageStore):
    """ For testing purpose """

    def store(self, msg):
        logger.debug("Should store message %s", msg)
        return 'fake_id'

    def get(self, id):
        return {'id':id, 'state': 'processed', 'message': None}

    def search(self, order_by='timestamp'):
        return []


class MemoryMessageStoreFactory(MessageStoreFactory):
    """ Return a Memory message store """
    def __init__(self):
        self.base_dict = {}

    def get_store(self, store_id):
        return MemoryMessageStore(self.base_dict, store_id)


class MemoryMessageStore(MessageStore):
    """ Store messages in memory """

    def __init__(self, base_dict, store_id):
        super().__init__()
        self.messages = base_dict.setdefault(store_id, OrderedDict())

    def store(self, msg):
        msg_id = msg.uuid.hex
        self.messages[msg_id] = {'id': msg_id, 'state': Message.PENDING, 'message': msg.to_dict()}
        return msg_id

    def change_message_state(self, id, new_state):
        self.messages[id]['state'] = new_state

    def get(self, id):
        resp = dict(self.messages[id])
        resp['message'] = Message.from_dict(resp['message'])
        return resp

    def search(self, order_by='timestamp'):

        for value in self.messages.values():

            resp = dict(value)
            resp['message'] = Message.from_dict(resp['message'])

            yield resp


class FileMessageStoreFactory(MessageStoreFactory):
    """ Return an FileMessageStore message store instance"""

    def __init__(self, path):
        super().__init__()
        self.base_path = path

    def get_store(self, store_id):
        return FileMessageStore(self.base_path, store_id)


class FileMessageStore(MessageStore):
    """ Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy."""

    def __init__(self, path, store_id):
        super().__init__()
        self.base_path = os.path.join(path, store_id)

        # Match msg file name
        self.msg_re = re.compile(r'^([0-9]{8})_([0-9]{2})([0-9]{2})_[0-9abcdef]*$')

    def store(self, msg):
        """ Store a file in `<base_path>/<store_id>/<month>/<day>/` hierachy."""

        # The filename is the file id
        filename = "{}_{}".format(msg.timestamp.strftime(DATE_FORMAT), msg.uuid.hex)
        dirs = os.path.join(str(msg.timestamp.year), "%02d" % msg.timestamp.month, "%02d" % msg.timestamp.day)

        try:
            # Try to make dirs if necessary
            os.makedirs(os.path.join(self.base_path, dirs))
        except FileExistsError:
            pass

        file_path = os.path.join(dirs, filename)

        # Write message to file
        with open(os.path.join(self.base_path, file_path), "w") as f:
            f.write(msg.to_json())

        self.change_message_state(file_path, Message.PENDING)

        return file_path

    def change_message_state(self, id, new_state):
        with open(os.path.join(self.base_path, id + '.meta'), "w") as f:
            f.write(new_state)

    def get_message_state(self, id):
        with open(os.path.join(self.base_path, id + '.meta'), "r") as f:
            state = f.read()
            return state

    def get(self, id):
        if not os.path.exists(os.path.join(self.base_path, id)):
            raise IndexError

        with open(os.path.join(self.base_path, id), "rb") as f:
            msg = Message.from_json(f.read().decode('utf-8'))
            return {'id':id, 'state': self.get_message_state(id), 'message': msg}

    def sorted_list_directories(self, path, reverse=True):
        """
        :param path: Base path
        :param reverse: reverse order
        :return: List of directories in specified path ordered
        """
        return sorted([d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))], reverse=reverse)

    def search(self, order_by='timestamp'):
        reverse = (order_by == 'timestamp')

        for year in self.sorted_list_directories(os.path.join(self.base_path), reverse=reverse):
            for month in self.sorted_list_directories(os.path.join(self.base_path, year), reverse=reverse):
                for day in self.sorted_list_directories(os.path.join(self.base_path, year, month), reverse=reverse):
                    for msg_name in sorted(os.listdir(os.path.join(self.base_path, year, month, day)), reverse=reverse):
                        found = self.msg_re.match(msg_name)
                        if found:
                            id = os.path.join(year, month, day, msg_name)
                            yield self.get(id)



import os
import json
import types
import asyncio
import logging

from datetime import datetime
from collections import OrderedDict

from urllib import parse

from concurrent.futures import ThreadPoolExecutor

from pypeman.message import Message
from pypeman.channels import Dropped, Break

logger = logging.getLogger(__name__)
loop = asyncio.get_event_loop()

from copy import deepcopy

# All declared nodes register here
all = []

# used to share external dependencies
ext = {}


class BaseNode:
    """ Base of all Nodes.
    If you create a new node, you must inherit from this class and implement `process` method.
    """
    dependencies = []

    def __init__(self, *args, **kwargs):
        self.channel = None
        all.append(self)
        self.name = kwargs.pop('name',self.__class__.__name__ + "_" + str(len(all)))
        self.store_output_as = kwargs.pop('store_output_as', None)
        self.store_input_as = kwargs.pop('store_input_as', None)
        self.passthrough = kwargs.pop('passthrough', None)
        self.next_node = None

    def requirements(self):
        """ List dependencies of modules if any """
        return self.dependencies

    def import_modules(self):
        """ Use this method to import specific external modules listed in dependencies """
        pass

    @asyncio.coroutine
    def handle(self, msg):
        """ Handle message is called by channel to launch process method on it.
        Some other structural processing take place here.
        Please, don't modify unless you know what you are doing.

        :param msg: incoming message
        :return: modified message after a process call and some treatment
        """

        # TODO : Make sure exceptions are well raised (does not happen if i.e 1/0 here atm)
        if self.store_input_as:
            msg.ctx[self.store_input_as] = dict(
                meta=dict(msg.meta),
                payload=deepcopy(msg.payload),
            )

        if self.passthrough:
            old_msg = msg.copy()

        result = self.run(msg)

        if isinstance(result, asyncio.Future):
            result = yield from result


        if self.next_node:
            if isinstance(result, types.GeneratorType):
                for res in result:
                    result = yield from self.next_node.handle(res)
                    # TODO Here result is last value returned. Is it a good idea ?
            else:
                if self.store_output_as:
                    result.ctx[self.store_output_as] = dict(
                        meta=dict(result.meta),
                        payload=deepcopy(result.payload),
                    )

                if self.passthrough:
                    result.payload = old_msg.payload
                    result.meta = old_msg.meta

                result = yield from self.next_node.handle(result)

        return result

    def run(self, msg):
        """ Used to overload behaviour like thread Node without rewriting handle process """
        result = self.process(msg)
        return result

    def process(self, msg):
        """ Implement this function in child classes to create
        a new Node.
        :param msg: The incoming message
        :return: The processed message
        """
        return msg

    def __str__(self):
        return "<%s(%s)>" % (self.channel.name, self.name)

class RaiseError(BaseNode):
    def process(self, msg):
        raise Exception("Test node")


class DropNode(BaseNode):
    """ This node used to tell the channel the message is Dropped. """
    def process(self, msg):
        raise Dropped()


class BreakNode(BaseNode):
    """ This node used to tell the channel the message is ????. """
    def process(self, msg):
        raise Break()


class Empty(BaseNode):
    """ Return an empty new message. """
    def process(self, msg):
        return Message()


class SetCtx(BaseNode):
    """ Push the message in the context with the key `ctx_name` """

    def __init__(self, ctx_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ctx_name = ctx_name

    def process(self, msg):
        msg.meta = msg.ctx[self.ctx_name]['meta']
        msg.payload = msg.ctx[self.ctx_name]['payload']

        return msg

class ThreadNode(BaseNode):
    """ Inherit from this class instead of BaseNode to avoid
    long run node blocking main event loop.
    """
    # TODO create class ThreadPool or channel ThreadPool or Global ?

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.executor = ThreadPoolExecutor(max_workers=3)

    def run(self, msg):
        result = self.channel.loop.run_in_executor(self.executor, self.process, msg)

        return result


class Log(BaseNode):
    """ Node to show some information about node, channel and message. Use for debug.
    """
    def __init__(self, *args, **kwargs):
        self.lvl = kwargs.pop('level', logging.INFO)
        self.show_ctx = kwargs.pop('show_ctx', None)
        super().__init__(*args, **kwargs)

    def process(self, msg):
        self.channel.logger.log(self.lvl, '%s %s', repr(self), repr(msg))

        if self.channel.parent_uids:
            self.channel.logger.log(self.lvl, 'Parent channels: %r', repr(self.channel.parent_names))

        self.channel.logger.log(self.lvl, 'Payload: %r', repr(msg.payload))

        if self.show_ctx:
            self.channel.logger.log(self.lvl, 'Contexts: %r', [repr(ctx) for ctx in msg.ctx])

        return msg

class JsonToPython(BaseNode):
    """ Convert json message payload to python dict."""
    # TODO encoding management
    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = json.loads(msg.payload, encoding=self.encoding)
        msg.content_type = 'application/python'
        return msg


class PythonToJson(BaseNode):
    """ Convert python payload to json."""
    def process(self, msg):
        msg.payload = json.dumps(msg.payload)
        msg.content_type = 'application/json'
        return msg


class XMLToPython(BaseNode):
    """ Convert XML message payload to python dict."""

    dependencies = ['xmltodict']

    def __init__(self, *args, **kwargs):
        self.process_namespaces = kwargs.pop('process_namespaces', False)
        super().__init__(*args, **kwargs)

    def import_modules(self):
        if 'xmltodict' not in ext:
            import xmltodict
            ext['xmltodict'] = xmltodict

    def process(self, msg):
        msg.payload = ext['xmltodict'].parse(msg.payload, process_namespaces=self.process_namespaces)
        msg.content_type = 'application/python'
        return msg


class PythonToXML(BaseNode):
    """ Convert python payload to XML."""
    dependencies = ['xmltodict']

    def __init__(self, *args, **kwargs):
        self.pretty = kwargs.pop('pretty', False)
        super().__init__(*args, **kwargs)

    def import_modules(self):
        if 'xmltodict' not in ext:
            import xmltodict
            ext['xmltodict'] = xmltodict

    def process(self, msg):
        msg.payload = ext['xmltodict'].unparse(msg.payload, pretty=self.pretty)
        msg.content_type = 'application/xml'
        return msg


class Encode(BaseNode):
    """ Encode payload in specified encoding to byte.
    """
    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = msg.payload.encode(self.encoding)
        return msg


class Decode(BaseNode):
    """ Decode payload from byte to specified encoding
    """
    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        super().__init__(*args, **kwargs)

    def process(self, msg):
        msg.payload = msg.payload.decode(self.encoding)
        return msg


# TODO put stores in specific file ?
class NullStoreBackend():
    """ For testing purpose """
    def store(self, message):
        pass


class FileStoreBackend():
    """ Backend used to store message with `MessageStore` node.
    """
    def __init__(self, path, filename, channel):
        self.path = path
        self.filename = filename
        self.counter = 0
        self.channel = channel

    def store(self, message):
        today = datetime.now()

        context = {'counter':self.counter,
                   'year': today.year,
                   'month': today.month,
                   'day': today.day,
                   'hour': today.hour,
                   'second': today.second,
                   'muid': message.uuid,
                   'cuid': getattr(self.channel, 'uuid', '???')
                   }

        filepath = os.path.join(self.path, self.filename % context)

        try:
            # Make missing dir if any
            os.makedirs(os.path.dirname(filepath))
        except FileExistsError:
            pass

        with open(filepath, 'w') as file_:
            file_.write(message.payload)

        self.counter += 1


class MessageStore(ThreadNode):
    """ Store a message in specified store """
    def __init__(self, *args, **kwargs):

        self.uri = kwargs.pop('uri')
        parsed = parse.urlparse(self.uri)

        super().__init__(*args, **kwargs)

        if parsed.scheme == 'file':
            filename = parsed.query.split('=')[1]

            self.backend = FileStoreBackend(path=parsed.path, filename=filename, channel=self.channel)
        else:
            self.backend = NullStoreBackend()


    def process(self, msg):
        self.backend.store(msg)
        return msg


class HL7ToPython(BaseNode):
    """ Convert hl7 payload to python struct."""

    dependencies = ['hl7']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def import_modules(self):
        if 'hl7' not in ext:
            import hl7
            ext['hl7'] = hl7

    def process(self, msg):
        msg.payload = ext['hl7'].parse(msg.payload)
        msg.content_type = 'application/python'
        return msg


class PythonToHL7(BaseNode):
    """ Convert python payload to HL7."""

    dependencies = ['hl7']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def import_modules(self):
        if 'hl7' not in ext:
            import hl7
            ext['hl7'] = hl7

    def process(self, msg):
        msg.payload = str(msg.payload)
        msg.content_type = 'text/hl7'
        return msg


class FileReader(BaseNode):
    """ Reads a file and sets payload to the file's contents. """
    def __init__(self, filename=None, path=None, binary_file=False, *args, **kwargs):
        self.filename = filename
        self.path = path
        self.binary_file = binary_file
        self.counter = 0
        super().__init__(*args, **kwargs)

    def process(self, msg):
        if self.filename:
            if callable(self.filename):
                name = self.filename(msg)
            else:
                name = self.filename
        else:
            name = msg.meta['filename']

        if self.path:
            path = self.path
        else:
            path = os.path.dirname(msg.meta['filepath'])

        filepath = os.path.join(path, name)

        if self.binary_file:
            mode = "rb"
        else:
            mode = "r"
        with open(filepath, mode) as file:
            msg.payload = file.read()
            msg.meta['filename'] = name
            msg.meta['filepath'] = filepath

        self.counter += 1
        return msg


class FileWriter(BaseNode):
    """ Write a file with the message content. """
    def __init__(self, filename=None, path=None, binary_mode=False, safe_file=False, *args, **kwargs):
        self.filename = filename
        self.path = path
        self.binary_mode = binary_mode
        self.counter = 0
        self.safe_file = safe_file
        super().__init__(*args, **kwargs)

    def process(self, msg):

        if self.filename:
            name = self.filename
        else:
            name = msg.meta['filename']

        if self.path:
            path = self.path
        else:
            path = os.path.dirname(msg.meta['filepath'])

        today = datetime.now()

        context = {'counter': self.counter,
                   'year': today.year,
                   'month': today.month,
                   'day': today.day,
                   'hour': today.hour,
                   'second': today.second,
                   }

        dest = os.path.join(path, name % context)

        old_file = dest
        if self.safe_file:
            dest = old_file + '.tmp'

        with open(dest, 'w' + ('b' if self.binary_mode else '')) as file_:
            file_.write(msg.payload)

        if self.safe_file:
            os.rename(dest, old_file)

        self.counter += 1

        return msg


class MappingNode(BaseNode):
    """ Used to map input message keys->values to another keys->values """

    def __init__(self, *args, **kwargs):
        self.mapping = kwargs.pop('mapping')
        self.recopy = kwargs.pop('recopy')
        path = kwargs.pop('path', "")

        self.path = 'payload'

        if path:
            self.path += '.' + path

        super().__init__(*args, **kwargs)

    def process(self, msg):
        current = msg
        parts = self.path.split('.')
        for part in parts:
            try:
                current = current[part]
            except (TypeError, KeyError):
                current = getattr(current, part)

        old_dict = current
        new_dict = {}

        for mapItem in self.mapping:
            mapItem.conv(old_dict, new_dict, msg)
        if self.recopy:
            new_dict.update(old_dict)

        dest = msg
        for part in parts[:-1]:
            try:
                dest = dest[part]
            except (TypeError, KeyError):
                dest = getattr(dest, part)

        try:
            dest[parts[-1]] = new_dict
        except KeyError:
            setattr(dest, parts[-1], new_dict)

        return msg


class RequestNode(ThreadNode):
    """ Http request node """
    dependencies = ['requests']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = kwargs.pop('url')
        self.content_type = kwargs.pop('content_type', None)
        self.auth = kwargs.pop('auth', None)
        self.verify = kwargs.pop('verify', False)
        self.url = self.url.replace('%(meta.', '%(')
        self.payload_in_url_dict = 'payload.' in self.url

        # TODO: create used payload keys for better perf of generate_request_url()

    def import_modules(self):
        """ import modules """
        if 'requests' not in ext:
            import requests
            ext['requests'] = requests

    def generate_request_url(self, msg):

        url_dict = msg.meta
        if self.payload_in_url_dict:
            url_dict = dict(url_dict)
            try:
                for key, val in msg.payload.items():
                    url_dict['payload.' + key] = val
            except AttributeError:
                logger.exception("Payload must be a python dict if used to generate url. This can be fixed using JsonToPython node before your RequestNode")
                raise

        logger.debug("Completing url %r with data from %r" % (self.url, url_dict))
        return self.url % url_dict

    def handle_request(self, msg):
        """ generate url and handle request """
        url = self.generate_request_url(msg)
        logger.debug("Destination request url: %s", url)
        resp = ext['requests'].get(url=url, auth=self.auth, verify=self.verify)
        return str(resp.text)

    def process(self, msg):
        """ handles request """
        msg.payload = self.handle_request(msg)

        return msg


class ToOrderedDict(BaseNode):
    """ this node yields an ordered dict with the keys 'keys' and the values from the payload
       if the payload does not contain certain values defaults can be specified with defaults
    """
    NONE = object()
    def __init__(self, *args, **kwargs):
        self.keys = kwargs.pop('keys')
        defaults = kwargs.pop('defaults', dict())
        self.dflt_dict = OrderedDict()
        for key in self.keys:
            self.dflt_dict[key] = defaults.get(key, ToOrderedDict.NONE)
        path = kwargs.pop('path', None)
        self.path = 'payload'
        if path:
            self.path += '.' + path
        super().__init__(*args, **kwargs)


    def process(self, msg):
        current = msg
        parts = self.path.split('.')

        for part in parts:
            try:
                current = current[part]
            except (TypeError, KeyError):
                current = getattr(current, part)
        old_dict = current
        new_dict = OrderedDict()

        for key in self.keys:
            val = old_dict.get(key, ToOrderedDict.NONE)
            if val is ToOrderedDict.NONE:
                val = self.dflt_dict[key]
            if val is not ToOrderedDict.NONE:
                new_dict[key] = val

        dest = msg
        for part in parts[:-1]:
            try:
                dest = dest[part]
            except (TypeError, KeyError):
                dest = getattr(dest, part)

        try:
            dest[parts[-1]] = new_dict
        except KeyError:
            setattr(dest, parts[-1], new_dict)

        return msg



#!/usr/bin/env python
import datetime
import uuid
import copy
from uuid import UUID
import pickle
import json
import base64


DATE_FORMAT = '%Y%m%d_%H%M'

class Message():
    # TODO : add_ctx and delete_ctx

    PENDING = "pending"
    PROCESSING = "processing"
    ERROR = "error"
    REJECTED = "rejected"
    PROCESSED = "processed"

    def __init__(self, content_type='application/text', payload=None, meta=None):
        self.content_type = content_type
        self.timestamp = datetime.datetime.now()
        self.uuid = uuid.uuid4()

        self.payload = payload

        if meta is None:
            meta = {}
        self.meta = meta
        
        self.ctx = {}
        
    def copy(self):
        """
        Copy the message. Useful for channel fork purpose.
        :return:
        """
        return copy.deepcopy(self)

    def to_dict(self):
        """
        Convert a message object to a dict.
        :return: A dict with an equivalent of message
        """
        result = {}
        result['timestamp'] = self.timestamp.strftime(DATE_FORMAT)
        result['uuid'] = self.uuid.hex
        result['payload'] = base64.b64encode(pickle.dumps(self.payload)).decode('ascii')
        result['meta'] = self.meta
        result['ctx'] = {}

        for k, m in self.ctx.items():
            result['ctx'][k] = m.to_dict()

        return result

    def to_json(self):
        return json.dumps(self.to_dict())

    @staticmethod
    def from_dict(data):
        """
        Convert the input dict previously converted with `.as_dict()` method in Message object.
        :param data: The input dict
        :return: A message object
        """
        result = Message()
        result.timestamp = datetime.datetime.strptime(data['timestamp'], DATE_FORMAT)
        result.uuid = UUID(data['uuid'])
        result.payload = pickle.loads(base64.b64decode(data['payload'].encode('ascii')))
        result.meta = data['meta']

        for k, m in data['ctx'].items():
            result.ctx[k] = Message.from_dict(m)

        return result

    @staticmethod
    def from_json(data):
        msg = Message.from_dict(json.loads(data))
        return msg

    def __str__(self):
        return "<msg: %s>" % self.uuid



#!/usr/bin/env python

# ############################################################################
# This is the code for THE pypeman command line program
# and all it's related commands
# after installing the module pypeman, all this functionality can be obtained 
# by calling 'pypeman' from the command line
# ############################################################################

import os
import sys


# Keep this import
sys.path.insert(0, os.getcwd())

import asyncio
import traceback
import importlib
import begin
import warnings
from functools import partial

from pypeman.helpers.reloader import reloader_opt
from pypeman import channels
from pypeman import nodes
from pypeman import endpoints

def load_project():
    try:
        importlib.import_module('project')
    except ImportError as exc:
        msg = str(exc)
        if not 'No module' in msg:
            print("IMPORT ERROR project")
            raise
        if not 'project' in msg:
            print("IMPORT ERROR project")
            raise
    except:
        import traceback

        traceback.print_exc()
        raise


def main(debug_asyncio=False, profile=False, cli=False):
    print('\nStart...')


    load_project()

    loop = asyncio.get_event_loop()

    if debug_asyncio:
        loop.set_debug(True)
        warnings.simplefilter('default')

    # Import modules for endpoints
    for end in endpoints.all:
        end.import_modules()

    # Import modules for nodes
    for node in nodes.all:
        node.import_modules()

    # Start channels
    for chan in channels.all:
        chan.import_modules()
        loop.run_until_complete(chan.start())

    # And endpoints
    for end in endpoints.all:
        loop.run_until_complete(end.start())

    # At the moment miing the asyncio and ipython
    # event loop is not working properly.
    # Thus ipython
    if cli:
        from pypeman.helpers.cli import CLI
        namespace = dict(
            loop=loop,
            nodes=nodes,
            endpoints=endpoints,
            channels=channels,
            )
        cli = CLI(namespace=namespace)
        cli.run_as_thread()

    print('Waiting for messages...')
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    print("End started tasks...")

    for chan in channels.all:
        loop.run_until_complete(chan.stop())

    pending = asyncio.Task.all_tasks()
    loop.run_until_complete(asyncio.gather(*pending))

    loop.close()


@begin.subcommand
def start(reload: 'Make server autoreload (Dev only)'=False, 
        debug_asyncio: 'Enable asyncio debug'=False,
        cli : "enables an IPython CLI for debugging (not perational)"=False,
        profile : "enables profiling / run stats (not operational)"=False,
        ):
    """ Start pypeman """
    reloader_opt(partial(main, debug_asyncio=debug_asyncio, cli=cli, profile=profile), reload, 2)


@begin.subcommand
def graph(dot: "Make dot compatible output"=False):
    """ Show channel graph """

    load_project()

    if dot:
        print("digraph testgraph{")
        for channel in channels.all:
            if not issubclass(channel.__class__, channels.SubChannel) and not issubclass(channel.__class__, channels.ConditionSubChannel):
                print("{node[shape=box]; %s; }" % channel.name)
                print(channel.name, end='')
                channel.graph_dot(previous=channel.name, end=channel.name)
        print("}")
    else:
        for channel in channels.all:
            if not issubclass(channel.__class__, channels.SubChannel) and not issubclass(channel.__class__, channels.ConditionSubChannel):
                print(channel.__class__.__name__)
                channel.graph()
                print('|-> out')
                print()


@begin.subcommand
def requirements():
    """ List optional python dependencies """

    load_project()

    dep = set()

    for channel in channels.all:
        dep |= set(channel.requirements())

    for node in nodes.all:
        dep |= set(node.requirements())

    for end in endpoints.all:
        dep |= set(end.requirements())

    [print(d) for d in dep]

@begin.subcommand
def startproject(dirname : "name of dir to install project to"):
    """ Creates a pypeman project from scrach """
    from pypeman.pjt_templates import new_project
    new_project(dirname)


@begin.subcommand
def debug():
    """ Used for development purpose """
    pass


@begin.start
def run(test=False, 

        ):
    """ Pypeman is a minimalistic but pragmatic ESB/ETL in python """



from setuptools import setup
from importlib import import_module

VERSION = import_module("pypeman").__version__
URL = import_module("pypeman").__url__

setup(name='pypeman',
      version=VERSION,
      description='Minimalistic but pragmatic ESB / ETL / EAI in Python',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.4',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Healthcare Industry',
        'Operating System :: POSIX :: Linux',
        'Topic :: Database',
        'Topic :: Internet',
        'Topic :: Internet :: File Transfer Protocol (FTP)',
        'Programming Language :: Python',
      ],
      keywords='esb etl eai pipeline data processing asyncio http',
      url=URL,

      author='Jeremie Pardou',
      author_email='jeremie.pardou@mhcomm.fr',

      license='Apache Software License',
      packages=['pypeman', 'pypeman.helpers', 'pypeman.tst_helpers'],
      entry_points = { 'console_scripts': [
            'pypeman = pypeman.commands:run.start',
        ]},

      test_suite='nose.collector',
      install_requires=['begins'],
      tests_require=['nose', 'nose-cover3']
)




import asyncio
import datetime
import os
import sys
import uuid
import logging
import re
import types
import warnings


from pypeman import endpoints, message, msgstore

logger = logging.getLogger(__name__)

# List all channel registered
all = []

# used to share external dependencies
ext = {}


class Dropped(Exception):
    """ Used to stop process as message is unusefull. Default success should be returned.
    """
    pass


class Rejected(Exception):
    """ Used to tell caller the message is invalid with a error return.
    """
    pass


class Break(Exception):
    """ Used to break message processing and return default success.
    """
    pass


class ChannelStopped(Exception):
    pass


class BaseChannel:
    STARTING, WAITING, PROCESSING, STOPPING, STOPPED  = range(5)

    dependencies = [] # List of module requirements

    def __init__(self, name=None, parent_channel=None, loop=None, force_msg_order=True, message_store_factory=None):
        self.uuid = uuid.uuid4()

        all.append(self)
        self._nodes = []
        self.status = None

        if name:
            self.name = name
        else:
            warnings.warn("Channels without names are deprecated", DeprecationWarning)
            self.name = self.__class__.__name__ + "_" + str(len(all))

        if loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = loop

        self.logger = logging.getLogger('pypeman.channels')

        if parent_channel:
            self.parent_uids = [parent_channel.uuid]
            self.parent_names = [parent_channel.name]
            if parent_channel.parent_uids:
                self.parent_uids.append(parent_channel.parent_uids)
                self.parent_names.append(parent_channel.parent_names)
        else:
            self.parent_uids = None

        self.next_node = None

        message_store_factory = message_store_factory or msgstore.NullMessageStoreFactory()

        self.message_store = message_store_factory.get_store(self.name)

        # Used to avoid multiple messages processing at same time
        self.lock = asyncio.Lock(loop=self.loop)


    def requirements(self):
        """ List dependencies of modules if any """
        return self.dependencies

    def import_modules(self):
        """ Use this method to import specific external modules listed in dependencies """
        pass

    @asyncio.coroutine
    def start(self):
        """ Start the channel """
        self.status = BaseChannel.STARTING
        self.init_node_graph()
        self.status = BaseChannel.WAITING

    def init_node_graph(self):
        if self._nodes:
            previous_node = self._nodes[0]

            for node in self._nodes[1:]:
                previous_node.next_node = node
                previous_node = node

    @asyncio.coroutine
    def stop(self):
        """ Stop the channel """
        self.status = BaseChannel.STOPPING
        # TODO Verify that all messages are processed
        self.status = BaseChannel.STOPPED

    def add(self, *args):
        """
        Add specified nodes to channel.
        :param args: Nodes to add
        :return: -
        """
        for node in args:
            node.channel = self
            self._nodes.append(node)
        return self

    def fork(self):
        """
        Create a new channel that process a copy of the message at this point.
        :return: The forked channel
        """
        s = SubChannel(parent_channel=self, loop=self.loop)
        self._nodes.append(s)
        return s

    def when(self, condition):
        """
        New channel bifurcation that is executed only if condition is True.
        :param condition: Can be a value or a function with a message argument.
        :return: The conditionnal path channel.
        """
        s = ConditionSubChannel(condition, parent_channel=self, loop=self.loop)
        self._nodes.append(s)
        return s

    @asyncio.coroutine
    def handle(self, msg):
        """ Overload this method only if you know what you are doing.
        :param msg: To be processed msg.
        :return: Processed message
        """

        if self.status in [BaseChannel.STOPPED, BaseChannel.STOPPING]:
            raise ChannelStopped

        self.logger.info("%s handle %s", self, msg)

        # Store message before any processing
        # TODO If store fails, do we stop processing ?
        msg_store_id = self.message_store.store(msg)

        # Only one message at time
        # TODO use keep_order var
        with (yield from self.lock):
            self.status = BaseChannel.PROCESSING
            try:
                result = yield from self.subhandle(msg)
                self.message_store.change_message_state(msg_store_id, message.Message.PROCESSED)
                return result
            except Dropped:
                self.message_store.change_message_state(msg_store_id, message.Message.PROCESSED)
                raise
            except Rejected:
                self.message_store.change_message_state(msg_store_id, message.Message.REJECTED)
                raise
            except Break:
                self.message_store.change_message_state(msg_store_id, message.Message.PROCESSED)
                raise
            except:
                self.logger.exception('Error while processing message %s', msg)
                self.message_store.change_message_state(msg_store_id, message.Message.ERROR)
                raise
            finally:
                self.status = BaseChannel.WAITING

    @asyncio.coroutine
    def subhandle(self, msg):
        """ Overload this method only if you know what you are doing.
        :param msg: To be processed msg.
        :return: Processed message
        """

        result = yield from self.process(msg)

        if self.next_node:
            if isinstance(result, types.GeneratorType):
                for res in result:
                    result = yield from self.next_node.handle(res)
                    # TODO Here result is last value returned. Is it a good idea ?
            else:
                result = yield from self.next_node.handle(result)

        return result

    @asyncio.coroutine
    def process(self, msg):
        """ Overload this method only if you know what you are doing.
        :param msg: To be processed msg.
        :return: Processed message
        """

        if self._nodes:
            res = yield from self._nodes[0].handle(msg)
            return res
        else:
            return msg

    def graph(self, prefix='', dot=False):
        for node in self._nodes:
            if isinstance(node, SubChannel):
                print(prefix + '|—\\')
                node.graph(prefix= '|  ' + prefix)
            elif isinstance(node, ConditionSubChannel):
                print(prefix + '|?\\')
                node.graph(prefix='|  ' + prefix)
                print(prefix + '|  -> Out')
            else:
                print(prefix + '|-' + node.name)

    def graph_dot(self, previous='', end=''):
        after = []
        for node in self._nodes:
            if isinstance(node, SubChannel):
                after.append((previous, '', node))
            elif isinstance(node, ConditionSubChannel):
                after.append((previous, end, node))
            else:
                print('->' + node.name, end='')
                previous = node.name
        if end:
            print("->" + end + ";")
        else:
            print(";")

        for prev, end, sub in after:
            print(prev, end='')
            sub.graph_dot(previous=prev, end=end)

    def __str__(self):
        return "<chan: %s>" % self.name


class SubChannel(BaseChannel):
    """ Subchannel used for fork """

    @asyncio.coroutine
    def process(self, msg):
        if self._nodes:
            asyncio.async(self._nodes[0].handle(msg.copy()), loop=self.loop)

        return msg


class ConditionSubChannel(BaseChannel):
    """ ConditionSubchannel used for make alternative path but join at the end """

    def __init__(self, condition, **kwargs):
        super().__init__(**kwargs)
        self.condition = condition

    def test_condition(self, msg):
        if callable(self.condition):
            return self.condition(msg)
        return True

    @asyncio.coroutine
    def subhandle(self, msg):
        if self.test_condition(msg):
            result = yield from self.process(msg)
        else:
            if self.next_node:
                result = yield from self.next_node.handle(msg)
            else:
                result = msg

        return result


class CaseSubChannel(BaseChannel):
    """ CaseSubchannel used for make alternative path without join at the end """

    def __init__(self, condition, **kwargs):
        super().__init__(**kwargs)
        self.condition = condition

    def test_condition(self, msg):
        if callable(self.condition):
            return self.condition(msg)
        return True

    @asyncio.coroutine
    def process(self, msg):
        if self.test_condition(msg):
            result = yield from self._nodes[0].handle(msg)
            return result

        return msg


class HttpChannel(BaseChannel):
    dependencies = ['aiohttp']
    app = None

    def __init__(self, *args, endpoint=None, method='*', url='/', encoding=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.method = method
        self.url = url
        self.encoding = encoding
        if endpoint is None:
            raise TypeError('Missing "endpoint" argument')
        self.http_endpoint = endpoint

    def import_modules(self):
        if 'aiohttp_web' not in ext:
            from aiohttp import web
            ext['aiohttp_web'] = web

    @asyncio.coroutine
    def start(self):
        yield from super().start()
        self.http_endpoint.add_route(self.method, self.url, self.handle)

    @asyncio.coroutine
    def handle(self, request):
        content = yield from request.text()
        msg = message.Message(content_type='http_request', payload=content, meta={'method': request.method})
        try:
            result = yield from super().handle(msg)
            encoding = self.encoding or 'utf-8'
            return ext['aiohttp_web'].Response(body=result.payload.encode(encoding), status=result.meta.get('status', 200))

        except Dropped:
            return ext['aiohttp_web'].Response(body="Dropped".encode('utf-8'), status=200)
        except Exception as e:
            return ext['aiohttp_web'].Response(body=str(e).encode('utf-8'), status=503)


class FileWatcherChannel(BaseChannel):
    NEW, UNCHANGED, MODIFIED, DELETED  = range(4)

    def __init__(self, *args, path='', regex='.*', interval=1, binary_file=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.path = path
        self.regex = regex
        self.interval = interval
        self.dirflag = os.path.isdir(self.path)
        self.data = {}
        self.re = re.compile(self.regex)
        self.binary_file = binary_file

        # Set mtime for all existing matching files
        if os.path.exists(self.path):
            for filename in os.listdir(self.path):
                if self.re.match(filename):
                    filepath = os.path.join(self.path, filename)
                    mtime = os.stat(filepath).st_mtime
                    self.data[filename] = mtime
        else:
            self.logger.warning('path not exist: %r', self.path)

    @asyncio.coroutine
    def start(self):
        yield from super().start()
        asyncio.async(self.watch_for_file(), loop=self.loop)

    def file_status(self, filename):
        if filename in self.data:
            old_mtime = self.data[filename]
            filepath = os.path.join(self.path, filename)
            new_mtime = os.stat(filepath).st_mtime
            if new_mtime == old_mtime:
                return FileWatcherChannel.UNCHANGED
            elif new_mtime > old_mtime:
                return FileWatcherChannel.MODIFIED
        else:
            return FileWatcherChannel.NEW

    def watch_for_file(self):
        yield from asyncio.sleep(self.interval)
        try:
            if os.path.exists(self.path):
                listfile = os.listdir(self.path)
                listfile.sort()

                for filename in listfile:
                    if self.re.match(filename):
                        status = self.file_status(filename)
                        # TODO watch deleted files ?
                        if status in [FileWatcherChannel.MODIFIED, FileWatcherChannel.NEW]:
                            filepath = os.path.join(self.path, filename)
                            self.data[filename] =  os.stat(filepath).st_mtime

                            # Read file and make message
                            if self.binary_file:
                                mode = "rb"
                            else:
                                mode = "r"

                            with open(filepath, mode) as file:
                                msg = message.Message()
                                msg.payload = file.read()
                                msg.meta['filename'] = filename
                                msg.meta['filepath'] = filepath
                                asyncio.async(super().handle(msg))

        finally:
            if not self.status in (BaseChannel.STOPPING, BaseChannel.STOPPED,):
                asyncio.async(self.watch_for_file(), loop=self.loop)


class TimeChannel(BaseChannel):
    dependencies = ['aiocron']

    def __init__(self, *args, cron='', **kwargs):
        super().__init__(*args, **kwargs)
        self.cron = cron

    def import_modules(self):
        if 'aiocron_crontab' not in ext:
            from aiocron import crontab

            ext['aiocron_crontab'] = crontab

    @asyncio.coroutine
    def start(self):
        super().start()
        ext['aiocron_crontab'](self.cron, func=self.tic, start=True)

    @asyncio.coroutine
    def tic(self):
        msg = message.Message()
        msg.payload = datetime.datetime.now()
        yield from self.handle(msg)


    @asyncio.coroutine
    def handle(self, msg):
        result = yield from super().handle(msg)
        return result


class MLLPChannel(BaseChannel):
    dependencies = ['hl7']

    def __init__(self, *args, endpoint=None, encoding='utf-8', **kwargs):
        super().__init__(*args, **kwargs)
        if endpoint is None:
            raise TypeError('Missing "endpoint" argument')
        self.mllp_endpoint = endpoint

        if encoding is None:
            encoding = sys.getdefaultencoding()
        self.encoding = encoding

    def import_modules(self):
        if 'hl7' not in ext:
            import hl7
            ext['hl7'] = hl7

    @asyncio.coroutine
    def start(self):
        yield from super().start()
        self.mllp_endpoint.set_handler(handler=self.handle)

    @asyncio.coroutine
    def handle(self, hl7_message):
        content = hl7_message.decode(self.encoding)
        msg = message.Message(content_type='text/hl7', payload=content, meta={})
        try:
            result = yield from super().handle(msg)
            return result.payload.encode(self.encoding)
        except Dropped:
            ack = ext['hl7'].parse(content, encoding=self.encoding)
            return str(ack.create_ack('AA')).encode(self.encoding)
        except Rejected:
            ack = ext['hl7'].parse(content, encoding=self.encoding)
            return str(ack.create_ack('AR')).encode(self.encoding)
        except Exception:
            ack = ext['hl7'].parse(content, encoding=self.encoding)
            return str(ack.create_ack('AE')).encode(self.encoding)
